# -*- coding:utf-8 -*-
# CREATED BY: bohuai jiang 
# CREATED ON: 2020/8/14 4:35 PM
# LAST MODIFIED ON: 2020/10/27 7:04 PM
# AIM:
import attrdict

from .automata import condition, operation
from .automata import symbols

SYMBOLS = symbols.SYMBOLS


def init_nodes(**kwargs):
    max_len = kwargs.get('max_len', 128)
    hard_max = kwargs.get('hard_max', 300)
    min_len = kwargs.get('min_len', 15)

    # --- initialize condition & operation --- #
    edges = attrdict.AttrDict({
        'is_end_symbol': condition.IsEndSymbolZH(SYMBOLS),
        'is_bracket_close': condition.IsBracketClose(SYMBOLS),
        'is_book_close': condition.IsBookClose(SYMBOLS),
        'is_quote_close': condition.IsQuoteClose(SYMBOLS),
        'is_end_state': condition.IsEndState(SYMBOLS),
        'is_long_sentence': condition.IsLongSentence(SYMBOLS, max_len=max_len),
        'is_short_sentence': condition.IsShortSentence(SYMBOLS, min_len=min_len),
        # --- sepcial Condition --- #
        'is_right_quota': condition.IsRightQuotaZH(),
        'is_left_quota': condition.IsLeftQuotaZH(index=1),
        'is_leftQ_gt_1': condition.IsLeftQuotaGreaterThan(theta=1),
        'is_rightQ_before_leftQ': condition.IsRQuotaStickWithLQuota(),
        'is_empty': condition.IsSpace(SYMBOLS),
        'with_says': condition.WithSays(),
        'special_ends': condition.SpecialEnds()
    })

    nodes = attrdict.AttrDict({
        'do_cut': operation.Normal(SYMBOLS),
        'long_handler': operation.LongHandler(SYMBOLS, hard_max=hard_max),
        'short_handler': operation.ShortHandler(SYMBOLS),
        'init': operation.Indolent(SYMBOLS),
        'quota_handler': operation.Indolent(SYMBOLS),
        'end': operation.EndState(SYMBOLS)
    })
    return edges, nodes


# --- graph --- #
def long_short_cuter(**kwargs):
    edges, nodes = init_nodes(**kwargs)
    return {
        nodes.init: [
            {'edge': edges.is_end_state,
             'node': nodes.end},

            {'edge': edges.is_short_sentence,
             'node': nodes.init},

            {'edge': edges.is_long_sentence,
             'node': nodes.long_handler},

            {'edge': [edges.special_ends,
                      edges.is_leftQ_gt_1,
                      edges.is_rightQ_before_leftQ],
             'args': any,
             'node': nodes.do_cut},

            {'edge': edges.is_right_quota,
             'node': nodes.quota_handler},

            # -- normal cut -- #
            {'edge': [edges.is_end_symbol,
                      edges.is_bracket_close,
                      edges.is_book_close,
                      edges.is_quote_close],
             'args': all,
             'node': nodes.do_cut},

        ],
        nodes.quota_handler: [
            {'edge': edges.is_end_state,
             'node': nodes.end},

            {'edge': edges.is_left_quota,
             'node': nodes.do_cut},

            {'edge': edges.is_empty,
             'node': nodes.quota_handler},

            # -- normal cut -- #
            {'edge': [edges.is_end_symbol,
                      edges.is_bracket_close,
                      edges.is_book_close,
                      edges.is_quote_close],
             'args': all,
             'node': nodes.do_cut},

            {'edge': None,
             'node': nodes.init}
        ],
        nodes.do_cut: [
            {'edge': edges.is_end_state,
             'node': nodes.end},

            {'edge': None,  # else 状态
             'node': nodes.init},
        ],
        nodes.long_handler: [
            {'edge': edges.is_end_state,
             'node': nodes.end},

            {'edge': None,  # else 状态
             'node': nodes.do_cut},
        ]
    }


def simple_cuter(**kwargs):
    edges, nodes = init_nodes(**kwargs)
    return {
        nodes.init: [
            {'edge': edges.is_end_state,
             'node': nodes.end},

            {'edge': [edges.special_ends,
                      edges.is_leftQ_gt_1,
                      edges.is_rightQ_before_leftQ],
             'args': any,
             'node': nodes.do_cut},

            {'edge': edges.is_right_quota,
             'node': nodes.quota_handler},

            # -- normal cut -- #
            {'edge': [edges.is_end_symbol,
                      edges.is_bracket_close,
                      edges.is_book_close,
                      edges.is_quote_close],
             'args': all,
             'node': nodes.do_cut},

        ],
        nodes.quota_handler: [
            {'edge': edges.is_end_state,
             'node': nodes.end},

            {'edge': edges.is_left_quota,
             'node': nodes.do_cut},

            {'edge': edges.is_empty,
             'node': nodes.quota_handler},

            # -- normal cut -- #
            {'edge': [edges.is_end_symbol,
                      edges.is_bracket_close,
                      edges.is_book_close,
                      edges.is_quote_close],
             'args': all,
             'node': nodes.do_cut},

            {'edge': None,
             'node': nodes.init}
        ],
        nodes.do_cut: [
            {'edge': edges.is_end_state,
             'node': nodes.end},

            {'edge': None,  # else 状态
             'node': nodes.init},
        ],
    }



# -*- coding:utf-8 -*-
# CREATED BY: bohuai jiang 
# CREATED ON: 2020/8/12 4:42 PM
# LAST MODIFIED ON:
# AIM:
import re

SYMBOLS = {
    'quotation_left': re.compile(r'([“])'),
    'quotation_en': re.compile(r'(["])'),
    'quotation_right': re.compile(r'([”])'),
    'bracket_left': re.compile(r'([<{\[\(（【「])'),
    'bracket_right': re.compile(r'([\)\]}」】）>])'),
    'book_left': re.compile(r'([《])'),
    'book_right': re.compile(r'([》])'),
    'comma': re.compile(r'([,，;])'),
    'end_symbols': re.compile(r'([?\!…;？！。\.])'),
    'en_dot': re.compile(r'\.'),
    'all_symbols': re.compile(r'([\]\(【\?】。,！[”，\!\.「<>？"（“…《）}》」;:：\){\)\s])')
}


SYMBOLS_EN = {
    'quotation_left': re.compile(r'([“])'),
    's_quota_left': re.compile(r'([‘])'),
    'quotation_en': re.compile(r'(["])'),
    's_quota_en':  re.compile("(['])"),
    'quotation_right': re.compile(r'([”])'),
    's_quota_right': re.compile(r'([’])'),
    'all_s_quota': re.compile(r'([‘’\'])'),
    'all_quota': re.compile(r'([“"”])'),
    'bracket_left': re.compile(r'([<{\[\(（【「])'),
    'bracket_right': re.compile(r'([\)\]}」】）>])'),
    'book_left': re.compile(r'([《])'),
    'book_right': re.compile(r'([》])'),
    'comma': re.compile(r'([,，])'),
    'end_symbols': re.compile(r'([?\!…？！。\.])'),
    'en_dot': re.compile(r'\.'),
    'dash': re.compile(r'—'),
    'short_dash': re.compile(r'-'),
    'semicolon': re.compile(r'([;；])'),
    'all_symbols': re.compile(r'([\]\(【\?】。,！[”，\!\.「<>？"（“…《）}》」;；:：\'‘’\-\—\){\)\s])')
}



# -*- coding:utf-8 -*-
# CREATED BY: jiangbohuai
# CREATED ON: 2021/11/24 3:06 PM
# LAST MODIFIED ON:
# AIM:
import os
import re
import pathlib

from sentence_spliter.architect.graph_component import Graph, Operation
import sentence_spliter.zh_cutter.condition as cond
import sentence_spliter.zh_cutter.operation as opt
from sentence_spliter.zh_cutter.zh_sequence import Sequence, SequenceABC
from sentence_spliter.zh_cutter.symbol import Symbols
from sentence_spliter.utility.file_opt import read_file
from sentence_spliter.utility.trie import Trie

WEIGHT_PATH = os.path.join(str(pathlib.Path(__file__).absolute().parent), 'weights_list.txt')


def get_int_key(dictionary: dict):
    for key in dictionary.keys():
        if type(key) == int:
            return key
    else:
        return None


class IntegrityCut(Operation):
    def __init__(self, name: str = 'IntegrityCut', min_len: int = 6, max_len: int = 40):
        """
        尽量不在句内拆分, 切分引号啥的
        """
        super(IntegrityCut, self).__init__(name)
        # -- set operation -- #
        self.back_sentence_start = opt.BackToSentenceStart()
        # -- set condition -- #
        self.is_end_symbol = cond.IsEndSymbol()\
            .add_and(cond.IsNextLeftBracket(reverse=True))\
            .add_and(cond.IsBeforeQuote(reverse=True))\
            .add_or(cond.IsNextLeftQuota())\
            .add_or(cond.IsNextSymbolAfterQuotaLarger(length=4, min_length=3)) \
            .add_or(cond.IsRightBracket())
        self.not_in_whitelist = cond.TokenInWhiteList(reverse=True)
        self.is_all_close = cond.IsBracketClose().add_and(cond.IsQuotaClose()).add_and(cond.IsSingleQuotaClose())
        self.next_capital = cond.NextStartWithCapital().add_or(cond.NextStartWithNum())
        self.is_short_sentence = cond.IsShortSentence(min_length=min_len)
        self.is_sentence_dash = cond.IsSentenceDash()
        self.is_long_sentence = cond.IsLongSentence(max_length=max_len)

        self.right_quota = Symbols.quotation_right.value + Symbols.s_quota_right.value + \
                           Symbols.s_quota_en.value + Symbols.quotation_en.value
        self.is_not_comma_stick_quota = self.is_not_comma_stick_quota = cond.IsCommaStickWithQuota(
            name='NotCommaStickWithQuota', reverse=True)
        self.not_number_dot = cond.IsNumberDot(name='NotNumberDot', reverse=True)
        self.is_not_symbol = cond.IsNextSymbol(name='NotNextSymbol', reverse=True).add_or(cond.IsNextLeftQuota())
        self.is_not_colon_stick_quota = cond.IsColonStickWithQuota(name='NotColonStickWithQuota', reverse=True)

    def run(self, seq: Sequence) -> None:
        end_pos = seq.i
        self.back_sentence_start.run(seq)

        for idx, _ in enumerate(seq):
            if self.is_short_sentence(seq):
                continue
            if seq.i >= end_pos:
                break

            if self.is_end_symbol(seq) \
                    and self.is_all_close(seq) \
                    and self.is_not_comma_stick_quota(seq) \
                    and self.is_not_colon_stick_quota(seq) \
                    and self.not_in_whitelist(seq) \
                    and not seq[seq.i + 1] in self.right_quota \
                    and self.not_number_dot(seq) \
                    and self.is_not_symbol:
                return
            # -- ?'   ?) 这两个case
            if (seq.current_token in self.right_quota) or (seq.current_token in Symbols.bracket_right.value):
                if seq[seq.i - 1] in Symbols.end_symbols.value:  # 优先切这里
                    return
                elif seq[seq.i - 1] in Symbols.semicolon.value:  # 其次切这里
                    return

            # 有分号
            if seq.current_token in Symbols.semicolon.value:
                return
            # if self.is_sentence_dash(seq):
            #     return


class IntSentenceCut(Operation):
    def __init__(self, name: str = 'IntSentenceCut', min_len: int = 6, max_len: int = 40):
        """
        句子内部切分
        """
        super(IntSentenceCut, self).__init__(name)
        # -- set operation -- #
        self.back_sentence_start = opt.BackToSentenceStart()
        self.cut = opt.Cut()
        # -- set condition -- #
        self.is_quota_close = cond.IsQuotaClose().add_and(cond.IsSingleQuotaClose())
        self.is_bracket_close = cond.IsBracketClose()
        self.is_end_symbol = cond.IsEndSymbol().add_and(cond.IsNextLeftBracket(reverse=True))\
            .add_and(cond.IsBeforeQuote(reverse=True))\
            .add_and(cond.TokenInWhiteList(reverse=True))\
            .add_or(cond.IsNextLeftQuota())\
            .add_or(cond.IsNextSymbolAfterQuotaLarger(length=4, min_length=3)) \
            .add_or(cond.IsRightBracket())
        self.is_next_capital = cond.NextStartWithCapital().add_or(cond.NextStartWithNum())
        self.is_short_sentence = cond.IsShortSentence(min_length=min_len)

        self.is_right_quota = cond.IsRightQuota().add_or(cond.IsRightSingleQuota())
        self.is_blank = cond.IsBlank()
        self.max_len = max_len
        self.is_right_quota = cond.IsRightQuota().add_or(cond.IsRightSingleQuota())
        # -- load weight_list -- #
        self.weight = self.load_weight_list()

        self.all_quota = Symbols.all_quota.value + Symbols.all_s_quota.value
        self.is_not_symbol = cond.IsNextSymbol(name='NotNextSymbol', reverse=True).add_or(cond.IsNextLeftQuota())
        self.is_not_all_symbol = cond.IsNextAllSymbol(name='NotNextAllSymbol', reverse=True).add_or(
            cond.IsNextLeftQuota())
        # self.is_not_next_bracket = cond.IsNextLeftBracket(name='NotNextLeftBracket', reverse=True)

    @staticmethod
    def load_weight_list() -> Trie:
        tree = Trie('<root>')
        weight_list = [v for v in read_file(WEIGHT_PATH).splitlines() if v]
        for tokens in weight_list:
            tokens = tokens.split()
            tokens[-1] = int(tokens[-1])
            tree.add_phrase(tokens)
        return tree

    def run(self, seq: Sequence) -> None:
        max_score = -1
        best_i = -1
        length = seq.i - seq.sentence_start
        end_pos = seq.i
        half_len = max(length // 2, 1)
        self.back_sentence_start.run(seq)
        traverse_trie = False
        tree = self.weight
        split_i = seq.i
        weight = -1000
        n_word = 0
        for idx, _ in enumerate(seq):
            position_penalty = 1 - abs(idx - half_len) / half_len

            if seq.i >= end_pos:
                break

            if traverse_trie:
                score = get_int_key(tree.children)
                if score is not None:
                    weight += position_penalty
                    traverse_trie = False
                elif seq.current_token in tree:
                    tree = tree[seq.current_token]
                else:
                    tree = self.weight
                    traverse_trie = False
                    weight = -1000
            else:
                n_word = seq.n_words

            if self.is_blank(seq):
                continue
            # -- get comma
            if (seq.current_token in Symbols.comma.value and not seq[seq.i + 1] in self.all_quota) or \
                    (seq.current_token in Symbols.bracket_right.value and self.is_quota_close(seq) and
                     self.is_not_all_symbol(seq)) or \
                    (self.is_quota_close(seq) and self.is_bracket_close(seq) and self.is_right_quota(seq) and seq[
                        seq.i - 1] not in Symbols.comma.value) or \
                    seq.current_token in Symbols.colon.value:
                if not self.is_short_sentence(seq):
                    weight = 0.5 + position_penalty
                    split_i = seq.i
                    traverse_trie = True

            if self.is_end_symbol(seq) and self.is_not_symbol(seq) and 1 < n_word < self.max_len and not self.is_short_sentence(seq):
                score = 1
                if self.is_quota_close(seq) and self.is_bracket_close(seq):
                    score += 1
                weight = score + position_penalty
                split_i = seq.i

            if (self.is_right_quota(seq) or seq.current_token in Symbols.bracket_right.value) \
                    and 1 < n_word < self.max_len \
                    and seq.get_left_nearest_tok() not in Symbols.all_symbols() \
                    and re.match(r'^\s+$', seq[seq.i + 1]):
                score = 1
                if self.is_quota_close(seq) and self.is_bracket_close(seq):
                    score += 1
                weight = 1 + position_penalty
                split_i = seq.i

            if weight > max_score:
                best_i = split_i
                max_score = weight
        if best_i > 0:
            seq.i = best_i
            self.cut.run(seq)


class LongHandler(Operation):
    def __init__(self, name: str = 'LongHandler', min_len: int = 6, max_len: int = 40):
        super(LongHandler, self).__init__(name)
        # -- set operation -- #
        self.integrity_cut = IntegrityCut(min_len=min_len, max_len=max_len)
        self.in_sentence_cut = IntSentenceCut(max_len=max_len, min_len=min_len)
        self.cut = opt.Cut()
        # -- set condition -- #
        self.is_long_sentence = cond.IsLongSentence(max_length=max_len)

    def run(self, seq: Sequence) -> None:
        org_i = seq.i
        org_n_word = seq.n_words  # 防止死循环
        self.integrity_cut.run(seq)
        if org_i == seq.i:  # self.is_long_sentence(seq):
            self.in_sentence_cut.run(seq)
            # -- 防止死循环 -- #
            if org_i == seq.i:  # self.is_long_sentence(seq):
                seq.i = org_i
                seq.n_words = org_n_word
        else:
            return self.cut.run(seq)



# -*- coding:utf-8 -*-
# CREATED BY: jiangbohuai
# CREATED ON: 2021/11/11 11:04 AM
# LAST MODIFIED ON:
# AIM:
import re
from typing import List

from enum import Enum

regular_symbols = {
    '.': r'\.',
    '?': r'\?',
    '(': r'\(',
    ')': r'\)',
    '[': r'\[',
    ']': r'\]'
}


class Symbols(Enum):
    quotation_left = list('“')
    s_quota_left = list('‘')
    quotation_en = list('"')
    s_quota_en = list("'")
    quotation_right = list('”')
    s_quota_right = list('’')
    all_s_quota = list('‘’\'')
    all_quota = list('“"”')
    bracket_left = list('《<{[(（【「')
    bracket_right = list(')]}」】）>》')
    comma = list(',，')
    end_symbols = list('?!…？！。.')
    en_dot = list('.')
    dash = list('—')
    short_dash = list('-')
    all_dash = list('—-')
    semicolon = list(';；')
    colon = list('：:')

    @staticmethod
    def all_symbols():
        out = set()
        for v in Symbols:
            [out.add(vv) for vv in v.value]
        return ''.join(out)


# -*- coding:utf-8 -*-
# CREATED BY: bohuai jiang 
# CREATED ON: 2020/8/14 4:30 PM
# LAST MODIFIED ON: 2020/10/27 7:05 PM
# AIM:
from typing import List

from .automata.state_machine import StateMachine
from .automata.sequence import StrSequence, EnSequence

from .logic_graph import long_short_cuter, simple_cuter
from .logic_graph_en import long_cuter_en

# --  init default state machine -- #
__long_short_machine = StateMachine(long_short_cuter())
__simple_logic = StateMachine(simple_cuter())
__long_machine_en = StateMachine(long_cuter_en())


def cut_to_sentences(paragraph: str, long_short: bool = False, verbose: bool = False):
    m_input = StrSequence(paragraph, verbose)
    if long_short:
        __long_short_machine.run(m_input)
        return m_input.sentence_list()
    else:
        __simple_logic.run(m_input)
        return m_input.sentence_list()


def run_cut(str_block: str, logic_graph: dict) -> List[str]:
    machine = StateMachine(logic_graph)
    m_input = StrSequence(str_block)
    machine.run(m_input, verbose=False)
    return m_input.sentence_list()


def cut_to_sentences_en(paragraph: str, verbose: bool = False):
    m_input = EnSequence(paragraph, verbose)
    __long_machine_en.run(m_input)
    return m_input.sentence_list()



# -*- coding:utf-8 -*-
# CREATED BY: jiangbohuai
# CREATED ON: 2021/10/12 5:39 PM
# LAST MODIFIED ON:
# AIM:
from typing import List, Union
import abc

from loguru import logger

from sentence_spliter.architect.sequence import SequenceABC


class JudgeTree:
    """
    方便condition
    """

    def __init__(self, args: callable):
        self.next = None
        self.args = args
        self.condition = []
        self.parent = None

    def add_children(self, args: callable, cond: 'Condition'):
        if args == self.args:
            self.condition.append(cond)
            return self
        else:
            self.parent = JudgeTree(args)
            self.parent.condition.append(cond)
            self.parent.next = self
            return self.parent

    def get_root(self):
        root = self
        # --- find root -- #
        while root.next is not None:
            root = self.next
        return root

    def judge(self, tok: SequenceABC, result: bool) -> bool:
        root = self.get_root()

        while root is not None:
            result = self.args([v(tok) for v in self.condition] + [result])
            root = root.parent
        return result


class Condition:
    def __init__(self, name: str, reverse: bool = False):
        self.reverse = reverse  # reverse result
        self.name = name
        self.judge_tree = None
        self.allow_add = True

    @abc.abstractmethod
    def run(self, tok: SequenceABC) -> bool:
        pass

    def add_or(self, cond: 'Condition'):
        assert self.allow_add is True, f"{self} has already been call, add_or is disabled"
        if self.judge_tree is None:
            self.judge_tree = JudgeTree(args=any)
        self.judge_tree = self.judge_tree.add_children(args=any, cond=cond)
        return self

    def add_and(self, cond: 'Condition'):
        assert self.allow_add is True, f"{self} has already been call, add_and is disabled"
        if self.judge_tree is None:
            self.judge_tree = JudgeTree(args=all)
        self.judge_tree = self.judge_tree.add_children(args=all, cond=cond)
        return self

    def __str__(self):
        return self.name

    def __call__(self, tok: SequenceABC, debug: bool = False):

        if not self.reverse:
            out = self.run(tok)
        else:
            out = not self.run(tok)
        # --- run judge tree -- #
        if self.judge_tree is not None:
            self.judge_tree = self.judge_tree.get_root()
            out = self.judge_tree.judge(tok, out)
        self.allow_add = False
        if debug:
            logger.debug(f'\t - ({self.name}) - {out}')
        return out


class Operation:
    def __init__(self, name: str):
        self.children = []
        self.name = name

    def __str__(self):
        return self.name

    def add_child(self, node: 'Operation', edges: Union[List[Condition], Condition] = [], args: callable = all):
        if isinstance(edges, list):
            self.children.append((edges, node, args))
        else:
            self.children.append(([edges], node, args))
        return self

    @abc.abstractmethod
    def run(self, seq: SequenceABC) -> None:
        pass

    def __call__(self, seq: SequenceABC, debug: bool = False) -> 'Operation':
        self.run(seq)
        tok = seq
        for con, opt, args in self.children:
            if debug:
                args_str = 'all' if args == all else 'any'
                logger.debug(f'\t <{opt.name}>\\{args_str}')
            # -- debug here --#

            if args([v(tok, debug) for v in con]):
                return opt

        else:
            raise Exception('DEAD loop !!!')

    def is_dead_loop(self) -> bool:
        pass

    def render(self) -> bool:
        pass


class StateMachine:
    def __init__(self, init_state: Operation):
        self.vertex = init_state

    def process(self, seq: SequenceABC, debug: bool = True):
        # -- depth first -- #
        vertex = self.vertex
        if debug:
            logger.debug(f'<{vertex}> "{seq}"-[{seq.i}]')
        while vertex.name != 'EndState':
            # -- do operation -- #
            vertex = vertex(seq, debug)
            if debug:
                logger.debug(f'<{vertex}> "{seq}"-[{seq.i}]')
        vertex.run(seq)


class Graph(Operation):
    """
    里面存在子图
    """

    def __init__(self, name: str):
        super(Graph, self).__init__(name)
        self.init_conditions()
        self.init_operations()
        self.logic_tree = self.build_logic()

    @abc.abstractmethod
    def init_conditions(self):
        pass

    @abc.abstractmethod
    def init_operations(self):
        pass

    @abc.abstractmethod
    def build_logic(self) -> 'Operation':
        pass

    def run(self, seq: SequenceABC, debug: bool = False) -> None:
        machine = StateMachine(self.logic_tree)
        machine.process(seq, debug=debug)



# -*- coding:utf-8 -*-
# CREATED BY: jiangbohuai
# CREATED ON: 2021/1/25 5:43 PM
# LAST MODIFIED ON:
# AIM: 子句权重list


weights="""
, and 10
, or 10
, but 10
, even 10
, however 10
, whenever 10
, whatever 10
, although 10
, thought 10
, nevertheless 10
, because 10
, besides 10
, so that 10
, otherwise 10
, hence 10
, therefore 10
, meanwhile 10
, where 10
, what 10
, which 10
, when 10
, who 10
, whom 10
, thereby 10
, for 5
, to 5
, I 10
, you 10 
, he 10
, she 10
, there 5
, then 5
, that 5
, following 10
, followed 10
, after 5 
"""


# -*- coding:utf-8 -*-
# CREATED BY: jiangbohuai
# CREATED ON: 2021/11/15 4:09 PM
# LAST MODIFIED ON:
# AIM:
from sentence_spliter.architect.graph_component import Operation
from sentence_spliter.en_cutter.en_sequence import Sequence


class Indolent(Operation):
    def __init__(self, name: str = 'Indolent'):
        super(Indolent, self).__init__(name)

    def run(self, seq: Sequence) -> None:
        return


class EndState(Operation):
    def __init__(self, name: str = 'EndState', min_length: int = -1):
        super(EndState, self).__init__(name)
        self.min_len = min_length

    def run(self, seq: Sequence) -> None:
        # if seq.sentence_start != len(seq):
        #     if seq.sentence_list_idx:
        #         seq.add_to_results()
        #     else:
        #         last = (seq.sentence_start, len(seq))
        #         seq.sentence_list_idx.append(last)
        last_sentence_len = len(seq) - seq.sentence_start
        if seq.sentence_start != len(seq):
            if last_sentence_len > self.min_len or not seq.sentence_list_idx:
                seq.add_to_results()
            else:
                last = seq.sentence_list_idx.pop(-1)
                last = (last[0], len(seq))
                seq.sentence_list_idx.append(last)
        return


class Proceed(Operation):
    def __init__(self, name: str = 'Proceed'):
        super(Proceed, self).__init__(name)

    def run(self, seq: Sequence) -> None:
        seq.next()


class Cut(Operation):
    def __init__(self, name: str = 'Cut'):
        super(Cut, self).__init__(name)

    def run(self, seq: Sequence) -> None:
        return seq.add_to_results()


# ----- long short --- #

class BackToSentenceStart(Operation):
    def __init__(self, name: str = 'BackToSentenceStart'):
        super(BackToSentenceStart, self).__init__(name)

    def run(self, seq: Sequence) -> None:
        seq.i = seq.sentence_start


class ResetNumWords(Operation):
    def __init__(self, name: str = 'RestNumWords'):
        super(ResetNumWords, self).__init__(name)

    def run(self, seq: Sequence) -> None:
        seq.n_words -= 1
        # cnt = 1
        # while cnt > 0:
        #     seq.next()
        #     if seq.word_pattern.match(seq.current_token):
        #         cnt -= 1



# -*- coding:utf-8 -*-
# CREATED BY: jiangbohuai
# CREATED ON: 2021/11/23 10:26 AM
# LAST MODIFIED ON:
# AIM: 字典树

from typing import List, Optional


class Render:
    def __init__(self):
        self.render_list = []

    @staticmethod
    def child_mid() -> str:
        return '├──'

    @staticmethod
    def child_end() -> str:
        return '└──'

    @staticmethod
    def parent() -> str:
        return '│'


class Trie:
    def __init__(self, key: object):
        self.value = key
        self.children = dict()
        self.path = str(self.value)

    def __str__(self):
        return str(self.value)

    def __getitem__(self, item: str):
        return self.children[item]

    def __contains__(self, item: str):
        return item in self.children

    def __add_tok(self, tok: str):
        if tok not in self.children:
            child = Trie(tok)
            child.path = f'{self.path} {str(child)}'
            self.children[tok] = child
            return child
        return self.children[tok]

    def add_phrase(self, phrase: List[str]):
        tree = self
        for value in phrase:
            tree = tree.__add_tok(value)
        tree.__add_tok('<end>')
        return self

    def show(self, header: List = [], bool_last: bool = False):
        front = '\t'.join(header)
        print(f'{front}{self.value}')

        items_list = list(self.children.items())
        if not items_list:
            return
        elif header:
            if bool_last:
                header[-1] = ''
            else:
                header[-1] = Render.parent()
        for key, child in items_list[0:-1]:
            child.show(header + [Render.child_mid()])
        key, child = items_list[-1]
        child.show(header + [Render.child_end()], True)

    def next(self, tok: str):
        return self.children.get(tok, None)

    def is_end_node(self, tok:str):
        if tok not in self and '<end>' in self:
            return True
        return False


data = '''
Dr.
U!S!A!
No.
abbr.
Brig.
Ltd.,
b.
N.
hr.
Fr.
U.S.A.
Is.
Ms.
F.
N. 
Con.
Sun.
det.
Br.
cont.
cp.
Rep.
Dept.
Thur.
ref.
sq.
Cpl.
f.o.b.
ib.
inc.
par.
cert.
pron.
Co.
R.
Gt.
m.
I.
R.S.V.P.
sing.
ibid.
temp.
Maj.
Sep.
Tues.
Hon.
Pl.
anon.
Corp.
o.n.o.
W.
adj.
Dem.
Mar.
A.M.
Jun.
Rev.
O.
o.b.o.
adv.
Co.,
LL.D
Soc.
Aug.
Sat.
C.E.
So.
Y.
M.A.
exclam.
mm.
gm.
T.
enc.
G.
H.
B.Phil
C.
Jan.
Sec.
max.
U.S.
B.A.
Ct.
p.m.
tel.
n.
st.
N.Y.
prep.
M.S.
Cdr.
p.p.
lat.
Q.
B.S.
esp.
P.M.
Ind.
U. S.
Wed.
chap.
l.c.
doz.
R.I.P.
pl.
Secy.
A.D.
ft.
eng.
p.d.q.
ct.
Sq.
Capt.
q.v.
col.
D.O.B.
pop.
Lt.
Bros.
sec.
A.H.
L.
Av.
Lieut.
M.
u.c.
Mr.
conj.
Col.
r.h.
Mgr.
Inc.
incl.
yr.
P.
Washingto, D.C.
Assoc.
Oct.
neg.
S. 
ext.
Sgt.
govt.
encl.
H.E.
X.
M.O.
arr.
pt.
p.
Apr.
G.L.O.W.
a.m.
l.
H.M.
l.h.
f.
d.
O.R.P.
deg.
A.
Messrs.
H.K.
B.Litt
vol.
dd.
lb.
Asst.
c.
Esq.
Mon.
med.
Ft.
D.
D.C.
D.A.
Mt.
B.
Ave.
N.Z.
V.
Q.E.D.
S.E.
N.B.
Mrs.
B.C.E.
Univ.
lng.
p.w.
Blvd.
Jul.
min.
i.e.
E.
B.C.
C.-in-C.
U.
fig.
h.p.
Sr.
J.
e.g.
D.T.'s
gal.
Supt.
para.
B.Ed.
symb.
T.M.
Prof.
K.
LL.B
U.K.
P.E.
Op.
Gen.
L.A.
dep.
ff.
Ph.D.
D.Litt
do.
ed.
p.s.i.
Feb.
Sept.
viz.
Fri.
Z.
p.a.
Div.
cf.
S.
St.
cu.
attn.
Jr.
Lab.
Dec.
Rd.
d.b.a.
U.N.
Nov.
Sen.
cent.
LL.M
B.Mus
'''
